/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

/*
 * Map between ADIF modes & submodes used by ADIF 3.1.1 and our mode
 * strings.
 *
 * http://www.adif.org/311/ADIF_311.htm
 */

#include <jeffpc/types.h>

#include <hlog/adif.h>

struct mode {
	const char *mode;
	const char *adif_mode;
	const char *adif_submode;
	struct str adif_mode_val;
	struct str adif_submode_val;
};

#define MODE(hlog, amode, asub) \
	{ \
		.mode = (hlog), \
		.adif_mode = (amode), \
		.adif_submode = (asub), \
		.adif_mode_val = STR_STATIC_INITIALIZER(amode), \
		.adif_submode_val = STR_STATIC_INITIALIZER(asub), \
	}

/*
 * There are some questions to figure out answers to.  Until that time, we
 * don't provide a (hlog) mode name making the corresponding entry not used
 * for matching.
 *
 *  1. What is the best way to "collapse" and "expand" the more detailed
 *     (hlog) mode when converting?  For example, ADIF provides one
 *     mode/submode pair for DominoEX (i.e., DOMINO + DOMINOEX) but we track
 *     the variants (e.g., DOMEX8).  Converting from our mode to ADIF could
 *     be done with regular expressions or glob patterns, but what about
 *     converting from ADIF?  Do we simply generate a "bare" name?
 *
 *  2. What should be the (hlog) mode names for some of these modes?  For
 *     example, ADIF modes like PKT currently have no (hlog) mode name
 *     simply because that name hasn't be decided yet.
 *
 * Note that it is valid for an ADIF file to have a mode and no submode.
 * For example, it is valid to have a RTTY mode and no submode even though
 * the spec defines one (ASCII).  The below table makes these non-submode
 * mappings explicit as "!submode".  (Source: ADIF dev mailing list)
 */
static struct mode modes[] = {
	MODE("AM",                  "AM",                 NULL),
	MODE("ARDOP",               "ARDOP",              NULL),
	MODE("ATV",                 "ATV",                NULL),
	MODE("C4FM",                "C4FM",               NULL),
	MODE(NULL, /* ? */          "CHIP",               "CHIP64"),
	MODE(NULL, /* ? */          "CHIP",               "CHIP128"),
	MODE(NULL, /* ? */          "CLO",                NULL),
	MODE(NULL, /* ? */          "CONTESTI",           NULL),
	MODE(NULL, /* ? */          "CW",                 "PCW"),
	MODE(NULL, /* ? */          "DIGITALVOICE",       NULL),
	MODE(NULL, /* DOMEX* */     "DOMINO",             "DOMINOEX"),
	MODE("DOMF",                "DOMINO",             "DOMINOF"),
	MODE("DSTAR",               "DSTAR",              NULL),
	MODE("FAX",                 "FAX",                NULL),
	MODE("FM",                  "FM",                 NULL),
	MODE("FSK441",              "FSK441",             NULL),
	MODE("FT8",                 "FT8",                NULL),
	MODE(NULL, /* ? */          "HELL",               "FMHELL"),
	MODE(NULL, /* ? */          "HELL",               "FSKHELL"),
	MODE(NULL, /* ? */          "HELL",               "HELL80"),
	MODE(NULL, /* ? */          "HELL",               "HFSK"),
	MODE(NULL, /* ? */          "HELL",               "PSKHELL"),
	MODE(NULL, /* ? */          "ISCAT",              "ISCAT-A"),
	MODE(NULL, /* ? */          "ISCAT",              "ISCAT-B"),
	MODE("JT4", /* !submode */  "JT4",                NULL),
	MODE(NULL, /* ? */          "JT4",                "JT4A"),
	MODE(NULL, /* ? */          "JT4",                "JT4B"),
	MODE(NULL, /* ? */          "JT4",                "JT4C"),
	MODE(NULL, /* ? */          "JT4",                "JT4D"),
	MODE(NULL, /* ? */          "JT4",                "JT4E"),
	MODE(NULL, /* ? */          "JT4",                "JT4F"),
	MODE(NULL, /* ? */          "JT4",                "JT4G"),
	MODE(NULL, /* ? */          "JT6M",               NULL),
	MODE("JT9", /* !submode */  "JT9",                NULL),
	MODE(NULL, /* ? */          "JT9",                "JT9-1"),
	MODE(NULL, /* ? */          "JT9",                "JT9-2"),
	MODE(NULL, /* ? */          "JT9",                "JT9-5"),
	MODE(NULL, /* ? */          "JT9",                "JT9-10"),
	MODE(NULL, /* ? */          "JT9",                "JT9-30"),
	MODE(NULL, /* ? */          "JT9",                "JT9A"),
	MODE(NULL, /* ? */          "JT9",                "JT9B"),
	MODE(NULL, /* ? */          "JT9",                "JT9C"),
	MODE(NULL, /* ? */          "JT9",                "JT9D"),
	MODE(NULL, /* ? */          "JT9",                "JT9E"),
	MODE(NULL, /* ? */          "JT9",                "JT9E FAST"),
	MODE(NULL, /* ? */          "JT9",                "JT9F"),
	MODE(NULL, /* ? */          "JT9",                "JT9F FAST"),
	MODE(NULL, /* ? */          "JT9",                "JT9G"),
	MODE(NULL, /* ? */          "JT9",                "JT9G FAST"),
	MODE(NULL, /* ? */          "JT9",                "JT9H"),
	MODE(NULL, /* ? */          "JT9",                "JT9H FAST"),
	MODE(NULL, /* ? */          "JT44",               NULL),
	MODE("JT65", /* !submode */ "JT65",               NULL),
	MODE(NULL, /* ? */          "JT65",               "JT65A"),
	MODE(NULL, /* ? */          "JT65",               "JT65B"),
	MODE(NULL, /* ? */          "JT65",               "JT65B2"),
	MODE(NULL, /* ? */          "JT65",               "JT65C"),
	MODE(NULL, /* ? */          "JT65",               "JT65C2"),
	MODE(NULL, /* ? */          "MFSK",               "FSQCALL"),
	MODE("FST4",                "MFSK",               "FST4"),
	MODE("FT4",                 "MFSK",               "FT4"),
	MODE("JS8",                 "MFSK",               "JS8"),
	MODE("MFSK4",               "MFSK",               "MFSK4"),
	MODE("MFSK8",               "MFSK",               "MFSK8"),
	MODE("MFSK11",              "MFSK",               "MFSK11"),
	MODE("MFSK16",              "MFSK",               "MFSK16"),
	MODE("MFSK22",              "MFSK",               "MFSK22"),
	MODE("MFSK31",              "MFSK",               "MFSK31"),
	MODE("MFSK32",              "MFSK",               "MFSK32"),
	MODE("MFSK64",              "MFSK",               "MFSK64"),
	MODE("MFSK128",             "MFSK",               "MFSK128"),
	MODE("MSK144",              "MSK144",             NULL),
	MODE("MT63",                "MT63",               NULL),
	MODE("OL-4/125",            "OLIVIA",             "OLIVIA 4/125"),
	MODE("OL-4/250",            "OLIVIA",             "OLIVIA 4/250"),
	MODE("OL-8/250",            "OLIVIA",             "OLIVIA 8/250"),
	MODE("OL-8/500",            "OLIVIA",             "OLIVIA 8/500"),
	MODE("OL-16/500",           "OLIVIA",             "OLIVIA 16/500"),
	MODE("OL-16/1000",          "OLIVIA",             "OLIVIA 16/1000"),
	MODE("OL-32/1000",          "OLIVIA",             "OLIVIA 32/1000"),
	MODE(NULL, /* ? */          "OPERA",              "OPERA-BEACON"),
	MODE(NULL, /* ? */          "OPERA",              "OPERA-QSO"),
	MODE(NULL, /* ? */          "PAC",                "PAC2"),
	MODE(NULL, /* ? */          "PAC",                "PAC3"),
	MODE(NULL, /* ? */          "PAC",                "PAC4"),
	MODE(NULL, /* ? */          "PAX",                "PAX2"),
	MODE(NULL, /* ? */          "PKT",                NULL),
	MODE("FSK31",               "PSK",                "FSK31"),
	MODE("PSK10",               "PSK",                "PSK10"),
	MODE("PSK31",               "PSK",                "PSK31"),
	MODE("PSK63",               "PSK",                "PSK63"),
	MODE("PSK63F",              "PSK",                "PSK63F"),
	MODE("PSK125",              "PSK",                "PSK125"),
	MODE("PSK250",              "PSK",                "PSK250"),
	MODE("PSK500",              "PSK",                "PSK500"),
	MODE("PSK1000",             "PSK",                "PSK1000"),
	MODE("PSKAM10",             "PSK",                "PSKAM10"),
	MODE("PSKAM31",             "PSK",                "PSKAM31"),
	MODE("PSKAM50",             "PSK",                "PSKAM50"),
	MODE("PSKFEC31",            "PSK",                "PSKFEC31"),
	MODE("QPSK31",              "PSK",                "QPSK31"),
	MODE("QPSK63",              "PSK",                "QPSK63"),
	MODE("QPSK125",             "PSK",                "QPSK125"),
	MODE("QSPK250",             "PSK",                "QSPK250"),
	MODE("QPSK500",             "PSK",                "QPSK500"),
	MODE("SIM31",               "PSK",                "SIM31"),
	MODE("PSK2K",               "PSK2K",              NULL),
	MODE("Q15",                 "Q15",                NULL),
	MODE(NULL, /* ? */          "ORA64",              "QRA64A"),
	MODE(NULL, /* ? */          "ORA64",              "QRA64B"),
	MODE(NULL, /* ? */          "ORA64",              "QRA64C"),
	MODE(NULL, /* ? */          "ORA64",              "QRA64D"),
	MODE(NULL, /* ? */          "ORA64",              "QRA64E"),
	MODE(NULL, /* ? */          "ROS",                "ROS-EME"),
	MODE(NULL, /* ? */          "ROS",                "ROS-HF"),
	MODE(NULL, /* ? */          "ROS",                "ROS-MF"),
	MODE("RTTY", /* !submode*/  "RTTY",               NULL),
	MODE(NULL, /* ? */          "RTTY",               "ASCII"),
	MODE(NULL, /* ? */          "RTTYM",              NULL),
	MODE("SSB", /* !submode */  "SSB",                NULL),
	MODE("LSB",                 "SSB",                "LSB"),
	MODE("USB",                 "SSB",                "USB"),
	MODE("SSTV",                "SSTV",               NULL),
	MODE("T10",                 "T10",                NULL),
	MODE(NULL, /* THOR* */      "THOR",               NULL),
	MODE(NULL, /* ? */          "THRB",               "THRBX"),
	MODE(NULL, /* ? */          "TOR",                "AMTORFEC"),
	MODE(NULL, /* ? */          "TOR",                "GTOR"),
	MODE(NULL, /* ? */          "V4",                 NULL),
	MODE(NULL, /* ? */          "VOI",                NULL),
	MODE(NULL, /* ? */          "WINMOR",             NULL),
	MODE("WSPR",                "WSPR",               NULL),
};

static const struct mode *lookup_mode_by_hlog_mode(const char *hlog_mode)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(modes); i++) {
		const struct mode *mode = &modes[i];

		/* ignore entries without hlog mode name */
		if (!mode->mode)
			continue;

		if (strcasecmp(hlog_mode, mode->mode))
			continue;

		return mode;
	}

	return NULL;
}

static const struct mode *lookup_mode_by_adif_mode(const char *adif_mode,
						   const char *adif_submode)
{
	size_t i;

	for (i = 0; i < ARRAY_LEN(modes); i++) {
		const struct mode *mode = &modes[i];

		/* ignore entries without hlog mode name */
		if (!mode->mode)
			continue;

		if (strcasecmp(adif_mode, mode->adif_mode))
			continue;

		if (adif_submode &&
		    strcasecmp(adif_submode, mode->adif_submode))
			continue;

		return mode;
	}

	return NULL;
}

int adif_mode_lookup_by_hlog_mode(const char *hlog_mode, const char **adif_mode_r,
				  const char **adif_submode_r)
{
	const struct mode *mode;

	mode = lookup_mode_by_hlog_mode(hlog_mode);
	if (!mode)
		return -ENOENT;

	*adif_mode_r = mode->adif_mode;
	*adif_submode_r = mode->adif_submode;

	return 0;
}

struct str *adif_mode_lookup_mode_by_hlog_mode(const char *hlog_mode)
{
	const struct mode *mode;

	mode = lookup_mode_by_hlog_mode(hlog_mode);
	if (!mode)
		return ERR_PTR(-ENOENT);

	/* cast away const - this is safe since static vals don't change */
	return (struct str *) &mode->adif_mode_val;
}

struct str *adif_mode_lookup_submode_by_hlog_mode(const char *hlog_mode)
{
	const struct mode *mode;

	mode = lookup_mode_by_hlog_mode(hlog_mode);
	if (!mode)
		return ERR_PTR(-ENOENT);

	/* some modes in ADIF don't have a submode, for those we return NULL */
	if (!mode->adif_submode)
		return NULL;

	/* cast away const - this is safe since static vals don't change */
	return (struct str *) &mode->adif_submode_val;
}

const char *adif_mode_lookup_by_adif_mode(const char *adif_mode,
					  const char *adif_submode)
{
	const struct mode *mode;

	mode = lookup_mode_by_adif_mode(adif_mode, adif_submode);
	if (!mode)
		return ERR_PTR(-ENOENT);

	return mode->mode;
}
