/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#ifndef __QSO_H
#define __QSO_H

#include <jeffpc/error.h>
#include <jeffpc/int.h>
#include <jeffpc/val.h>
#include <jeffpc/uuid.h>
#include <jeffpc/nvl.h>
#include <jeffpc/refcnt.h>

#include <hlog/time.h>

struct qso_side {
	struct qso *qso; /* backpointer to the QSO */

	struct val *station_call;
	struct val *operator_call;
	struct val *power; /* mW */
	struct val *freq; /* Hz */
	struct val *mode;
	struct val *rst;
	struct val *name;
	struct val *qth;
	struct val *grid;
	struct val *country;
	struct val *state;
	struct val *county;
	struct val *continent;
	struct val *dxcc;
	struct val *itu;
	struct val *cqz;
	struct val *seq; /* contest qso seq # */
};

struct qso {
	struct xuuid uuid;
	char uuid_str[XUUID_PRINTABLE_STRING_LENGTH];

	struct time start;
	struct time end;
	struct val *comment;

	struct qso_side tx; /* local / tx */
	struct qso_side rx; /* remote / rx */

	/*
	 * additional key-value pairs from ADIF
	 *
	 * If an ADIF field doesn't fit anywhere else in this struct, we
	 * preserve it by storing the unknown fields' values in this nvlist
	 * and writing them out when rewriting the file.
	 */
	struct nvlist *additional;

	refcnt_t refcnt;
};

extern struct qso *qso_alloc(void);
extern void qso_free(struct qso *qso);
extern int qso_save(const struct qso *qso, int data_dir);
extern void qso_reset(struct qso *qso);
extern void qso_copy(struct qso *dst, struct qso *src);

REFCNT_INLINE_FXNS(struct qso, qso, refcnt, qso_free, NULL);

/*
 * Set the various parts of a QSO.  All passed in struct str * are consumed.
 */

/* per QSO */
extern int qso_set_datetime(struct qso *qso, bool start, const char *str);
extern int qso_set_datetime_val(struct qso *qso, bool start, struct val *val);
extern int qso_set_comment(struct qso *qso, const char *comment);
extern int qso_set_comment_str(struct qso *qso, struct str *comment);
extern int qso_set_comment_val(struct qso *qso, struct val *comment);

/* per side */
extern int qso_set_station_call(struct qso_side *side, const char *call);
extern int qso_set_station_call_val(struct qso_side *side, struct val *val);
extern int qso_set_operator_call(struct qso_side *side, const char *call);
extern int qso_set_operator_call_val(struct qso_side *side, struct val *val);
extern int qso_set_freq(struct qso_side *side, const char *str);
extern int qso_set_freq_val(struct qso_side *side, struct val *val);
extern int qso_set_grid(struct qso_side *side, const char *grid);
extern int qso_set_grid_val(struct qso_side *side, struct val *val);
extern int qso_set_power(struct qso_side *side, uint64_t milliwatts);
extern int qso_set_power_user(struct qso_side *side, const char *power);
extern int qso_set_power_val(struct qso_side *side, struct val *val);
extern int qso_set_mode(struct qso_side *side, const char *mode);
extern int qso_set_mode_val(struct qso_side *side, struct val *val);
extern int qso_set_rst(struct qso_side *side, const char *rst);
extern int qso_set_rst_val(struct qso_side *side, struct val *val);
extern int qso_set_name(struct qso_side *side, const char *name);
extern int qso_set_name_str(struct qso_side *side, struct str *name);
extern int qso_set_name_val(struct qso_side *side, struct val *val);
extern int qso_set_qth(struct qso_side *side, const char *qth);
extern int qso_set_qth_str(struct qso_side *side, struct str *qth);
extern int qso_set_qth_val(struct qso_side *side, struct val *val);
extern int qso_set_country(struct qso_side *side, const char *country);
extern int qso_set_country_str(struct qso_side *side, struct str *country);
extern int qso_set_country_val(struct qso_side *side, struct val *val);
extern int qso_set_state(struct qso_side *side, const char *state);
extern int qso_set_state_str(struct qso_side *side, struct str *state);
extern int qso_set_state_val(struct qso_side *side, struct val *val);
extern int qso_set_county(struct qso_side *side, const char *county);
extern int qso_set_county_str(struct qso_side *side, struct str *county);
extern int qso_set_county_val(struct qso_side *side, struct val *val);
extern int qso_set_continent(struct qso_side *side, const char *zone);
extern int qso_set_continent_val(struct qso_side *side, struct val *val);
/* valid DXCC entities are [0,inf) */
extern int qso_set_zone_dxcc(struct qso_side *side, uint64_t zone);
extern int qso_set_zone_dxcc_user(struct qso_side *side, const char *zone);
extern int qso_set_zone_dxcc_val(struct qso_side *side, struct val *val);
/* valid ITU zones are [1,90]; 0 clears the field */
extern int qso_set_zone_itu(struct qso_side *side, uint64_t zone);
extern int qso_set_zone_itu_user(struct qso_side *side, const char *str);
extern int qso_set_zone_itu_val(struct qso_side *side, struct val *val);
/* valid CQ zones are [1,40]; 0 clears the field */
extern int qso_set_zone_cqz(struct qso_side *side, uint64_t zone);
extern int qso_set_zone_cqz_user(struct qso_side *side, const char *str);
extern int qso_set_zone_cqz_val(struct qso_side *side, struct val *val);
extern int qso_set_seq(struct qso_side *side, uint64_t seq);
extern int qso_set_seq_user(struct qso_side *side, const char *seq);
extern int qso_set_seq_val(struct qso_side *side, struct val *val);

/*
 * Clear the various parts of a QSO.
 */

/* per QSO */
extern void qso_clear_datetime(struct qso *qso, bool start);
extern void qso_clear_comment(struct qso *qso);

/* per side */
extern void qso_clear_station_call(struct qso_side *side);
extern void qso_clear_operator_call(struct qso_side *side);
extern void qso_clear_freq(struct qso_side *side);
extern void qso_clear_grid(struct qso_side *side);
extern void qso_clear_power(struct qso_side *side);
extern void qso_clear_mode(struct qso_side *side);
extern void qso_clear_rst(struct qso_side *side);
extern void qso_clear_name(struct qso_side *side);
extern void qso_clear_qth(struct qso_side *side);
extern void qso_clear_country(struct qso_side *side);
extern void qso_clear_state(struct qso_side *side);
extern void qso_clear_county(struct qso_side *side);
extern void qso_clear_continent(struct qso_side *side);
extern void qso_clear_zone_dxcc(struct qso_side *side);
extern void qso_clear_zone_itu(struct qso_side *side);
extern void qso_clear_zone_cqz(struct qso_side *side);
extern void qso_clear_seq(struct qso_side *side);

/*
 * Get the various parts of a QSO.
 */

/* per QSO */
/* no qso_get_datetime */
extern struct str *qso_get_comment(const struct qso *qso);
extern struct val *qso_get_comment_val(const struct qso *qso);

/* per side */
extern struct str *qso_get_station_call(const struct qso_side *side);
extern struct val *qso_get_station_call_val(const struct qso_side *side);
extern struct str *qso_get_operator_call(const struct qso_side *side);
extern struct val *qso_get_operator_call_val(const struct qso_side *side);
extern struct val *qso_get_freq(const struct qso_side *side);
extern struct val *qso_get_band(const struct qso_side *side);
extern struct str *qso_get_grid(const struct qso_side *side);
extern struct val *qso_get_grid_val(const struct qso_side *side);
extern struct val *qso_get_power(const struct qso_side *side);
extern struct str *qso_get_mode(const struct qso_side *side);
extern struct val *qso_get_mode_val(const struct qso_side *side);
extern struct str *qso_get_rst(const struct qso_side *side);
extern struct val *qso_get_rst_val(const struct qso_side *side);
extern struct str *qso_get_name(const struct qso_side *side);
extern struct val *qso_get_name_val(const struct qso_side *side);
extern struct str *qso_get_qth(const struct qso_side *side);
extern struct val *qso_get_qth_val(const struct qso_side *side);
extern struct str *qso_get_country(const struct qso_side *side);
extern struct val *qso_get_country_val(const struct qso_side *side);
extern struct str *qso_get_state(const struct qso_side *side);
extern struct val *qso_get_state_val(const struct qso_side *side);
extern struct str *qso_get_county(const struct qso_side *side);
extern struct val *qso_get_county_val(const struct qso_side *side);
extern struct str *qso_get_continent(const struct qso_side *side);
extern struct val *qso_get_continent_val(const struct qso_side *side);
extern struct val *qso_get_zone_dxcc(const struct qso_side *side);
extern struct val *qso_get_zone_itu(const struct qso_side *side);
extern struct val *qso_get_zone_cqz(const struct qso_side *side);
extern struct val *qso_get_seq(const struct qso_side *side);

#endif
