/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/types.h>
#include <jeffpc/error.h>

#include <ctype.h>
#include <string.h>

#include <hlog/maidenhead.h>

struct grid {
	char start;
	char end;
	double step; /* degrees */
};

static const struct grid grids_lat[] = {
	{ 'A', 'R', 10. },	/* field */
	{ '0', '9', 1. },	/* square */
	{ 'a', 'x', 2.5/60. },	/* subsquare */
	{ '0', '9', 2.5/600. },	/* extended square */
};

static const struct grid grids_lon[] = {
	{ 'A', 'R', 20. },	/* field */
	{ '0', '9', 2. },	/* square */
	{ 'a', 'x', 5./60. },	/* subsquare */
	{ '0', '9', 5./600. },	/* extended square */
};

STATIC_ASSERT(sizeof(grids_lat) == sizeof(grids_lon));

static bool encode_grid(char *letter, double *val, const struct grid *g)
{
	unsigned int whole = *val / g->step;

	ASSERT3U(whole, <=, g->end - g->start);

	*letter = g->start + whole;
	*val -= whole * g->step;

	return true;
}

bool maidenhead_encode(char *grid, size_t gridsize, double lat, double lon)
{
	size_t i;

	if (!grid || (gridsize < 3))
		return false; /* need at least 2 chars + nul */
	if ((lat < -90) || (lat > +90))
		return false;
	if ((lon < -180) || (lon > +180))
		return false;

	lat += 90.;
	lon += 180.;

	for (i = 0; i < MIN((gridsize - 1) / 2, ARRAY_LEN(grids_lat)); i++) {
		if (!encode_grid(&grid[2 * i + 0], &lon, &grids_lon[i]))
			return false;
		if (!encode_grid(&grid[2 * i + 1], &lat, &grids_lat[i]))
			return false;
	}

	grid[2 * i] = '\0';

	return true;
}

static inline bool decode_grid(double *loc, double *last_step, char c,
			       const struct grid *g)
{
	/* uppercase input to match expected range */
	if (g->start == toupper(g->start))
		c = toupper(c);

	if ((c < g->start) || (c > g->end))
		return false;

	*loc += (c - g->start) * g->step;
	*last_step = g->step;

	return true;
}

bool maidenhead_decode(const char *grid, double *lat_r, double *lon_r)
{
	double lat, lon;
	double lat_step, lon_step;
	size_t len;
	size_t i;

	if (!grid || (strlen(grid) < 2))
		return false;

	len = strlen(grid);
	lat = -90.;
	lon = -180.;

	for (i = 0; i < MIN(len / 2, ARRAY_LEN(grids_lat)); i++) {
		if (!decode_grid(&lon, &lon_step, grid[2 * i + 0], &grids_lon[i]))
			return false;
		if (!decode_grid(&lat, &lat_step, grid[2 * i + 1], &grids_lat[i]))
			return false;
	}

	*lat_r = lat + lat_step / 2;
	*lon_r = lon + lon_step / 2;

	return true;
}
