/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <math.h>
#include <string.h>

#include <jeffpc/types.h>

#include <hlog/util.h>

static int extract_float(const char *input, int64_t *whole_r,
			 int64_t *frac_r, int *scale_r)
{
	uint32_t whole, frac;
	unsigned int scale;
	const char *dp;
	bool negative;
	int ret;

	/* negative sign */
	negative = (*input == '-');
	if ((*input == '+') || (*input == '-'))
		input++;

	/* find the decimal point */
	dp = strchr(input, '.');

	/* whole part */
	ret = str2u32_full(input, &whole, 10, dp ? '.' : '\0');
	if (ret)
		return ret;

	/* fractional part */
	if (dp) {
		ret = str2u32(dp + 1, &frac);
		if (ret)
			return ret;

		scale = pow(10, strlen(dp + 1));
	} else {
		frac = 0;
		scale = 1;
	}

	/* output */
	*whole_r = ((int64_t) whole) * (negative ? -1 : 1);
	*frac_r = ((int64_t) frac) * (negative ? -1 : 1);
	*scale_r = scale;

	return 0;
}

/*
 * Parse a float but return it scaled by the supplied scalar.
 *
 * For example:
 *   - "144.545" with a scale of 1000000 will return 144545000
 *   - "-1.5" with a scale of 3600 will return -5400
 */
int extract_float_as_int64(const char *input, unsigned int scale,
			   int64_t *result_r)
{
	int64_t whole, frac;
	int fscale;
	int ret;

	ret = extract_float(input, &whole, &frac, &fscale);
	if (ret)
		return ret;

	/* TODO: check for overflow */

	/* scale */
	*result_r = (whole * scale) + (frac * scale / fscale);

	return 0;
}

int extract_float_as_int32(const char *input, unsigned int scale,
			   int32_t *result_r)
{
	int64_t tmp;
	int ret;

	ret = extract_float_as_int64(input, scale, &tmp);
	if (ret)
		return ret;

	if ((tmp > INT32_MAX) || (tmp < INT32_MIN))
		return -ERANGE;

	*result_r = tmp;

	return 0;
}

int extract_float_as_double(const char *input, double *result_r)
{
	int64_t whole, frac;
	int scale;
	int ret;

	ret = extract_float(input, &whole, &frac, &scale);
	if (ret)
		return ret;

	/* output */
	*result_r = ((double) whole) + ((double) frac) / scale;

	return 0;
}

char *trim_ws(char *str)
{
	char *end;

	if (!str)
		return NULL;

	/* trim leading whitespace */
	str = find_nonws(str);

	/* find last non-whitespace char to keep */
	end = rfind_nonws(str);

	/* null terminate right after - trimming the rest */
	if (end[0] != '\0')
		end[1] = '\0';

	/* return new string */
	return str;
}

/* pretty-print the frequency, omitting trailing zeros as needed */
bool pretty_print_freq(char *buf, size_t bufsize, struct val *freq)
{
	uint64_t mhz, khz, hz;

	if (!freq) {
		buf[0] = '\0';
		return false; /* no freq */
	}

	hz = freq->i % 1000;
	khz = (freq->i / 1000) % 1000;
	mhz = freq->i / 1000000;

	/*
	 * Print the frequency in MHz as X.YYY or X.YYYZZZ depending on
	 * whether ZZZ equal zero.
	 */
	if (hz)
		snprintf(buf, bufsize, "%"PRIu64".%03"PRIu64"%03"PRIu64,
			 mhz, khz, hz);
	else
		snprintf(buf, bufsize, "%"PRIu64".%03"PRIu64, mhz, khz);

	return true;
}

/* pretty-print the power with appropriate units */
bool pretty_print_power(char *buf, size_t bufsize, struct val *power)
{
	uint64_t v = power ? power->i : 0;
	const char *unit;
	int width;

	if (!v) {
		buf[0] = '\0';
		return false; /* no power */
	}

	if (v < 1000) {
		unit = "mW";
		width = 3;
	} else {
		unit = "W";
		width = 4;

		/* round up */
		if ((v % 1000) >= 500)
			v += 1000;

		v /= 1000;
	}

	snprintf(buf, bufsize, "%*"PRIu64"%s", width, v, unit);

	return true;
}

bool pretty_print_float(char *buf, size_t bufsize, struct val *value,
			int scale, int digits)
{
	uint64_t whole;
	uint64_t frac;

	if (!value) {
		buf[0] = '\0';
		return false;
	}

	whole = value->i / scale;
	frac = value->i % scale;

	while (frac && ((frac % 10) == 0)) {
		frac /= 10;
		digits--;
	}

	if (frac)
		snprintf(buf, bufsize, "%"PRIu64".%0*"PRIu64, whole, digits,
			 frac);
	else
		snprintf(buf, bufsize, "%"PRIu64, whole);

	return true;
}
