/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <dirent.h>

#include <jeffpc/rbtree.h>
#include <jeffpc/buffer.h>
#include <jeffpc/synch.h>
#include <jeffpc/mem.h>
#include <jeffpc/io.h>

#include <hlog/time.h>
#include <hlog/bands.h>

#include <hlog-rpc/announce-pub.h>

#include "index.h"

struct call_entry;

struct index_entry {
	struct qso *qso;

	struct call_entry *station;
	struct call_entry *operator;

	struct str *fname;

	struct rb_node node_uuid;
	struct rb_node node_time;
};

struct call_entry {
	struct str *call;
	struct rb_node node;
	struct rb_tree entries;

	struct {
		struct rb_tree band_mode;
		struct rb_tree band;
		struct rb_tree mode;
	} stats;
};

struct call_entry_stat {
	struct index_stat stat;
	struct rb_node node;
};

struct call_entry_link {
	struct rb_node node;
	struct index_entry *entry;

	struct {
		struct call_entry_stat *band_mode;
		struct call_entry_stat *band;
		struct call_entry_stat *mode;
	} stats;
};

static LOCK_CLASS(index_lock_lc);
static struct lock index_lock;
static struct rb_tree qso_index_by_uuid; /* uuid */
static struct rb_tree qso_index_by_time; /* time, call, freq, mode, uuid */
static struct rb_tree call_index; /* call */

static int index_entry_cmp_uuid(const void *va, const void *vb)
{
	const struct index_entry *a = va;
	const struct index_entry *b = vb;

	return xuuid_compare(&a->qso->uuid, &b->qso->uuid);
}

static int index_entry_cmp_time(const void *va, const void *vb)
{
	const struct index_entry *a = va;
	const struct index_entry *b = vb;
	int ret;

	ret = time_cmp(&a->qso->start, &b->qso->start);
	if (ret)
		return ret;

	ret = time_cmp(&a->qso->end, &b->qso->end);
	if (ret)
		return ret;

	if (a->qso->rx.station_call && b->qso->rx.station_call) {
		ret = val_cmp(a->qso->rx.station_call,
			      b->qso->rx.station_call);
		if (ret)
			return ret;
	} else if (a->qso->rx.station_call) {
		return +1;
	} else if (b->qso->rx.station_call) {
		return -1;
	} else {
		/* both missing call - equal so far */
	}

	if (a->qso->tx.freq && b->qso->tx.freq) {
		if (a->qso->tx.freq < b->qso->tx.freq)
			return -1;
		if (a->qso->tx.freq > b->qso->tx.freq)
			return +1;
	} else if (a->qso->tx.freq) {
		return +1;
	} else if (b->qso->tx.freq) {
		return -1;
	} else {
		/* both missing freq - equal so far */
	}

	if (a->qso->tx.mode && b->qso->tx.mode) {
		ret = val_cmp(a->qso->tx.mode,
			      b->qso->tx.mode);
		if (ret)
			return ret;
	} else if (a->qso->tx.mode) {
		return +1;
	} else if (b->qso->tx.mode) {
		return -1;
	} else {
		/* both missing mode - equal so far */
	}

	/* finally, break any ties with the uuid */
	return index_entry_cmp_uuid(va, vb);
}

static int call_entry_cmp(const void *va, const void *vb)
{
	const struct call_entry *a = va;
	const struct call_entry *b = vb;

	return str_cmp(a->call, b->call);
}

static int call_entry_entries_cmp(const void *va, const void *vb)
{
	const struct call_entry_link *a = va;
	const struct call_entry_link *b = vb;

	return index_entry_cmp_time(a->entry, b->entry);
}

static int call_entry_stats_cmp(const void *va, const void *vb)
{
	const struct call_entry_stat *a = va;
	const struct call_entry_stat *b = vb;

	if (a->stat.band < b->stat.band)
		return -1;
	if (a->stat.band > b->stat.band)
		return +1;

	if (a->stat.mode == b->stat.mode)
		return 0;
	if (!a->stat.mode)
		return -1;
	if (!b->stat.mode)
		return +1;

	return str_cmp(a->stat.mode, b->stat.mode);
}

static void stats_init(struct rb_tree *tree)
{
	rb_create(tree, &call_entry_stats_cmp,
		  sizeof(struct call_entry_stat),
		  offsetof(struct call_entry_stat, node));
}

static struct call_entry_stat *stats_find(struct rb_tree *tree,
					  uint64_t band, struct str *mode,
					  bool alloc)
{
	struct call_entry_stat *stat;
	struct call_entry_stat key = {
		.stat.band = band,
		.stat.mode = mode,
	};

	stat = rb_find(tree, &key, NULL);
	if (stat || !alloc)
		return stat;

	stat = malloc(sizeof(struct call_entry_stat));
	if (!stat)
		return NULL;

	stat->stat.band = band;
	stat->stat.mode = str_getref(mode);
	stat->stat.count = 0;

	rb_add(tree, stat);

	return stat;
}

static struct index_entry *alloc_entry(const char *fname)
{
	struct index_entry *entry;

	entry = zalloc(sizeof(struct index_entry));
	if (!entry)
		return ERR_PTR(-ENOMEM);

	entry->fname = fname ? str_dup(fname) : NULL;
	if (IS_ERR(entry->fname))
		goto err;

	entry->qso = qso_alloc();
	if (!entry->qso)
		goto err_fname;

	return entry;

err_fname:
	str_putref(entry->fname);

err:
	free(entry);

	return ERR_PTR(-ENOMEM);
}

static void free_entry(struct index_entry *entry)
{
	qso_putref(entry->qso);
	str_putref(entry->fname);
	free(entry);
}

static struct call_entry *add_call(struct val **call_str)
{
	struct call_entry *call;
	struct call_entry key = {
		.call = val_cast_to_str(*call_str),
	};

	if (!*call_str)
		return NULL;

	call = rb_find(&call_index, &key, NULL);
	if (!call) {
		call = malloc(sizeof(struct call_entry));
		if (!call)
			return ERR_PTR(-ENOMEM);

		call->call = val_getref_str(*call_str);
		rb_create(&call->entries, &call_entry_entries_cmp,
			  sizeof(struct call_entry_link),
			  offsetof(struct call_entry_link, node));
		stats_init(&call->stats.band_mode);
		stats_init(&call->stats.band);
		stats_init(&call->stats.mode);

		rb_add(&call_index, call);
	}

	if (val_cast_to_str(*call_str) != call->call) {
		val_putref(*call_str);
		*call_str = str_getref_val(call->call);
	}

	return call;
}

static int link_call(struct call_entry *call, struct index_entry *entry)
{
	const uint64_t freq = entry->qso->tx.freq ? entry->qso->tx.freq->i : 0;
	struct call_entry_link *link;
	struct call_entry_stat *stat_band_mode;
	struct call_entry_stat *stat_band;
	struct call_entry_stat *stat_mode;
	uint64_t band;
	struct str *mode;

	band = bands_lookup_lambda_by_freq(freq);
	mode = val_cast_to_str(entry->qso->tx.mode);

	stat_band_mode = stats_find(&call->stats.band_mode, band, mode, true);
	stat_band = stats_find(&call->stats.band, band, mode, true);
	stat_mode = stats_find(&call->stats.mode, band, mode, true);

	link = malloc(sizeof(struct call_entry_link));

	if (!stat_band_mode || !stat_band || !stat_mode || !link)
		return -ENOMEM; /* stats are in rb trees, not leaked */

	link->entry = entry;
	link->stats.band_mode = stat_band_mode;
	link->stats.band = stat_band;
	link->stats.mode = stat_mode;

	rb_add(&call->entries, link);

	/* update the counts */
	stat_band_mode->stat.count++;
	stat_band->stat.count++;
	stat_mode->stat.count++;

	return 0;
}

static void unlink_call(struct call_entry *call, struct index_entry *entry)
{
	struct call_entry_link *link;
	struct call_entry_link key = {
		.entry = entry,
	};

	link = rb_find(&call->entries, &key, NULL);
	ASSERT3P(link, !=, NULL);

	rb_remove(&call->entries, link);

	link->stats.band_mode->stat.count--;
	link->stats.band->stat.count--;
	link->stats.mode->stat.count--;

	free(link);
}

static int add_entry_to_indices(struct index_entry *entry)
{
	struct index_entry *existing;
	int ret;

	/* first, insert it into the global/by-uuid index */
	existing = rb_insert(&qso_index_by_uuid, entry);
	if (existing)
		return -EEXIST;

	/*
	 * Since we have a unique uuid, we should never end up with a
	 * duplicate at this point.
	 */

	/* make sure all the relevant callsigns are indexed */
	entry->operator = add_call(&entry->qso->rx.operator_call);
	entry->station = add_call(&entry->qso->rx.station_call);

	if (entry->operator) {
		/* index by operator callsign */
		ret = link_call(entry->operator, entry);
		if (ret)
			return ret;
	}

	if (entry->station && (entry->station != entry->operator)) {
		/* index by station callsign */
		ret = link_call(entry->station, entry);
		if (ret)
			return ret;
	}

	/* finally, index it globally by time */
	rb_add(&qso_index_by_time, entry);

	return 0;
}

static void remove_entry_from_indices(struct index_entry *entry)
{
	rb_remove(&qso_index_by_uuid, entry);
	rb_remove(&qso_index_by_time, entry);

	if (entry->operator)
		unlink_call(entry->operator, entry);

	if (entry->station && (entry->station != entry->operator))
		unlink_call(entry->station, entry);
}

static int import_entry(int dirfd, const char *fname)
{
	struct index_entry *entry;
	struct stat statbuf;
	char *raw;
	int ret;

	entry = alloc_entry(fname);
	if (IS_ERR(entry->fname))
		return PTR_ERR(entry);

	raw = read_file_common(dirfd, fname, &statbuf);
	if (IS_ERR(raw)) {
		ret = PTR_ERR(raw);
		goto err;
	}

	ret = qso_unpack_hlq(entry->qso, raw, statbuf.st_size);

	free(raw);

	if (ret)
		goto err;

	MXLOCK(&index_lock);
	ret = add_entry_to_indices(entry);
	MXUNLOCK(&index_lock);

	if (ret)
		goto err;

	return 0;

err:
	free_entry(entry);

	return ret;
}

/* file name must be: <uuid>.hlq */
static bool verify_file_name(const char *fname)
{
	struct xuuid uuid;

	/* The -1 is because the define includes a byte for the nul */
	if (strlen(fname) != (XUUID_PRINTABLE_STRING_LENGTH - 1 + 4))
		return false;

	if (strcmp(fname + XUUID_PRINTABLE_STRING_LENGTH - 1, ".hlq"))
		return false;

	return xuuid_parse_no_nul(&uuid, fname);
}

static int import_all_entries(int qso_dir)
{
	struct dirent *de;
	DIR *dir;
	int ret;

	dir = fdopendir(qso_dir);
	if (!dir)
		return -errno;

	while ((de = readdir(dir))) {
		if (!strcmp(de->d_name, ".") ||
		    !strcmp(de->d_name, ".."))
			continue;

		/* only process files with file names of the form: <uuid>.hlq */
		if (!verify_file_name(de->d_name))
			continue;

		ret = import_entry(qso_dir, de->d_name);
		if (ret == -EEXIST)
			cmn_err(CE_WARN, "Failed to import entry %s - "
				"duplicate UUID\n", de->d_name);
		else if (ret)
			cmn_err(CE_WARN, "Failed to import entry %s - "
				"ignoring: %s\n", de->d_name, xstrerror(ret));
	}

	closedir(dir);

	return 0;
}

int index_init(void)
{
	uint64_t start_time;
	uint64_t end_time;
	int qso_dir;
	int ret;

	MXINIT(&index_lock, &index_lock_lc);
	rb_create(&qso_index_by_uuid, &index_entry_cmp_uuid,
		  sizeof(struct index_entry),
		  offsetof(struct index_entry, node_uuid));
	rb_create(&qso_index_by_time, &index_entry_cmp_time,
		  sizeof(struct index_entry),
		  offsetof(struct index_entry, node_time));
	rb_create(&call_index, &call_entry_cmp, sizeof(struct call_entry),
		  offsetof(struct call_entry, node));

	qso_dir = xopen("qso", O_RDONLY | O_DIRECTORY, 0);
	if (qso_dir < 0)
		return qso_dir;

	cmn_err(CE_INFO, "Loading existing QSOs...");

	start_time = gettime();
	ret = import_all_entries(qso_dir);
	end_time = gettime();

	if (ret) {
		cmn_err(CE_ERROR, "Error: failed to load existing entries: %s",
			xstrerror(ret));
	} else {
		double dt = (end_time - start_time) / 1e9;

		cmn_err(CE_INFO, "Loaded %zu QSOs in %.3f secs (%.3f QSOs/s)",
			rb_numnodes(&qso_index_by_uuid), dt,
			rb_numnodes(&qso_index_by_uuid) / dt);
	}

	return ret;
}

int index_get(const struct xuuid *uuid, struct qso *qso)
{
	struct qso keyqso = {
		.uuid = *uuid,
	};
	struct index_entry key = {
		.qso = &keyqso,
	};
	struct index_entry *entry;

	MXLOCK(&index_lock);

	entry = rb_find(&qso_index_by_uuid, &key, NULL);
	if (entry)
		qso_copy(qso, entry->qso);

	MXUNLOCK(&index_lock);

	return entry ? 0 : -ENOENT;
}

static void print_notification(enum announce_qso_msg_type type,
			       const char *source, const char *uuid,
			       struct str *call, struct str *mode,
			       struct str *grid)
{
	switch (type) {
		case ANNOUNCE_QSO_NEW:
		case ANNOUNCE_QSO_UPDATED: {
			const char *action = (type == ANNOUNCE_QSO_NEW) ?
				"added" : "updated";

			cmn_err(CE_INFO, "QSO %-7s %s %s %s %s (%s)",
				action, uuid, str_cstr(call), str_cstr(mode),
				str_cstr(grid), source);
			break;
		}
	}

	str_putref(call);
	str_putref(mode);
	str_putref(grid);
}

int index_save(struct qso *qso, const char *source)
{
	struct qso keyqso = {
		.uuid = qso->uuid,
	};
	struct index_entry key = {
		.qso = &keyqso,
	};
	enum announce_qso_msg_type ann;
	struct index_entry *entry;
	int ret;

	/* save the struct qso to disk */
	ret = qso_save(qso, AT_FDCWD);
	if (ret)
		return ret;

	/* update the index */
	MXLOCK(&index_lock);

	entry = rb_find(&qso_index_by_uuid, &key, NULL);
	if (!entry) {
		/* new entry */
		entry = alloc_entry(NULL);
		if (IS_ERR(entry)) {
			MXUNLOCK(&index_lock);
			ret = PTR_ERR(entry);
			goto err;
		}

		ann = ANNOUNCE_QSO_NEW;
	} else {
		/* update existing entry */
		ASSERT3P(entry->qso, !=, qso);
		ASSERT(!xuuid_compare(&entry->qso->uuid, &qso->uuid));

		/*
		 * TODO: we could try to figure out if we're changing an
		 * fields that mess with the ordering and skip the
		 * remove/add steps.
		 */
		remove_entry_from_indices(entry);

		ann = ANNOUNCE_QSO_UPDATED;
	}

	qso_copy(entry->qso, qso);

	ret = add_entry_to_indices(entry);
	ASSERT3S(ret, !=, -EEXIST);

	MXUNLOCK(&index_lock);

	if (ret)
		free_entry(entry);

err:
	if (ret) {
		cmn_err(CE_WARN, "Failed to update index after successful "
			"QSO write: %s", xstrerror(ret));
		cmn_err(CE_WARN, "Index may be inconsistent");
	} else {
		/*
		 * Let everyone know about the update
		 *
		 * Since we already dropped the index lock, it is possible
		 * that this notification will arrive after the just saved
		 * value has already been read by the client.  That's ok.
		 */
		print_notification(ann, source, qso->uuid_str,
				   qso_get_station_call(&qso->rx),
				   qso_get_mode(&qso->rx),
				   qso_get_grid(&qso->rx));
		announce_publish_qso(ann, &qso->uuid);
	}

	return ret;
}
