/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/mem.h>
#include <jeffpc/unicode.h>

#include <hlog/index.h>
#include <hlog/qso_line.h>

#include "recent.h"
#include "xcurses.h"

struct line {
	struct xuuid uuid;
	char str[];
};

static WINDOW *win;
static size_t nrows, ncols;
static bool reeval;
static struct line **lines;

int recent_init(int height, int width, int row, int col)
{
	size_t i;

	win = newwin(height, width, row, col);
	whrule(win, height - 1, width);

	nrows = height;
	ncols = width;
	lines = mem_reallocarray(NULL, nrows, sizeof(struct line));
	if (!lines)
		goto err;

	for (i = 0; i < nrows; i++) {
		lines[i] = zalloc(sizeof(struct line) + ncols + 1);
		if (!lines[i])
			goto err_free;
	}

	reeval = true;

	return 0;

err_free:
	for (i = 0; i < nrows; i++)
		free(lines[i]);

	free(lines);

err:
	delwin(win);

	return -ENOMEM;
}

void recent_destroy(void)
{
	size_t i;

	for (i = 0; i < nrows; i++)
		free(lines[i]);

	free(lines);

	delwin(win);
}

void recent_refresh(void)
{
	struct index_iter_ctx ctx;
	struct qso *qso;
	size_t i;

	if (!reeval)
		return;

	/* cache the text */
	for (qso = index_iter_history_start(&ctx, false), i = 0;
	     qso && (i < nrows);
	     qso = index_iter_history_next(&ctx), i++) {
		lines[i]->uuid = qso->uuid;
		qso_line_fmt(qso, lines[i]->str, ncols);
	}

	/* print cached text */
	for (i = 0; i < nrows; i++)
		mvwprintw(win, i, 0, "%-*s", ncols, lines[i]->str);

	wrefresh(win);
}

void recent_invalidate(void)
{
	reeval = true;
}

/*
 * Legacy recent code, to be removed in the future
 */
void refresh_recent_win(WINDOW *win, int height, int width, bool count)
{
	const int start_row = count ? 1 : 0;
	struct index_iter_ctx ctx;
	size_t total_qsos;
	struct qso *qso;
	int row;

	/* clear it */
	for (row = start_row; row < height; row++)
	     mvwprintw(win, row, 0, "%*s", width - 1, "");

	/* for each mode, print a line */
	for (qso = index_iter_history_start(&ctx, false), row = start_row;
	     qso && (row < height);
	     qso = index_iter_history_next(&ctx), row++) {
		char tmp[width];

		wmove(win, row, 0);

		qso_line_fmt(qso, tmp, sizeof(tmp));
		wprintw(win, "%s", tmp);
	}

	if (!count)
		return;

	/* print QSO count in header line */
	index_counts(&total_qsos);
	if (total_qsos) {
		char tmp[width];

		if (total_qsos > 1)
			snprintf(tmp, sizeof(tmp), "1…%d/%zu", row - 1,
				 total_qsos);
		else
			snprintf(tmp, sizeof(tmp), "1/1");

		/*
		 * HACK: count number of printable characters
		 */
		wmove(win, 0, width -
		      utf8_is_valid_str(tmp, strlen(tmp)) - 1);
		wprintw(win, "%s", tmp);
	}
}
