/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/error.h>
#include <jeffpc/mem.h>
#include <jeffpc/buffer.h>

#include "xform.h"

struct xfield {
	struct xfield_def def; /* the definition */
	struct xform *form;
	struct buffer *buf;

	struct {
		size_t off; /* offset into visible area */
		size_t scroll; /* scroll offset */
	} cursor;
};

static inline struct xfield *canonical_field(struct xform *form, int idx)
{
	ASSERT3S((idx), <, (form)->nfields);

	if (idx < 0)
		idx = form->curfield;

	return &form->fields[idx];
}

static void draw_field(struct xfield *field)
{
	const char *data;
	size_t size;

	data = buffer_data(field->buf);
	size = buffer_size(field->buf); /* FIXME: handle utf8 */

	/*
	 * Adjust the data & size to "scroll" the text within the field
	 */
	ASSERT3U(field->cursor.scroll, <=, size);

	if (field->cursor.scroll == size) {
		ASSERT0(size);
		data = "";
	} else {
		data += field->cursor.scroll;
		size -= field->cursor.scroll;
	}

	/*
	 * Adjust the size to show only as much fits in the field
	 */
	size = MIN(size, field->def.width);

	/*
	 * draw the field (underlined)
	 */
	wattrset(field->form->formwin, A_NORMAL | A_UNDERLINE);
	mvwprintw(field->form->formwin, field->def.row, field->def.col,
		  "%.*s", (int)size, data); /* draw content */
	mvwprintw(field->form->formwin, field->def.row, field->def.col + size,
		  "%*s", field->def.width - size, ""); /* draw padding */
	wattrset(field->form->formwin, A_NORMAL);
}

static void draw_fields(struct xform *form)
{
	size_t i;

	for (i = 0; i < form->nfields; i++)
		draw_field(&form->fields[i]);
}

static void draw_labels(struct xform *form)
{
	size_t i;

	for (i = 0; i < form->nlabels; i++) {
		const struct xfield_def *f = &form->labels[i].def;

		ASSERT3S(f->type, ==, XFIELD_TYPE_LABEL);

		wattrset(form->basewin, A_NORMAL | f->attr.color);
		mvwprintw(form->basewin, f->row, f->col, "%s",
			  str_cstr(f->label));
		wattrset(form->basewin, A_NORMAL);
	}
}

int xform_alloc(struct xform *form, int height, int width, int row,
		int col,
		const struct xfield_def *fields, size_t nfields)
{
	size_t fi;
	size_t li;
	size_t i;
	int ret;

	form->basewin = newwin(height, width, row, col);
	form->formwin = derwin(form->basewin, height, width, 0, 0);

	form->nfields = 0;
	form->nlabels = 0;

	/* count fields & labels */
	for (i = 0; i < nfields; i++) {
		switch (fields[i].type) {
			case XFIELD_TYPE_FIELD:
				form->nfields++;

				/* fields have no labels of their own */
				if (fields[i].label) {
					ret = -EINVAL;
					goto err;
				}
				break;
			case XFIELD_TYPE_LABEL:
				form->nlabels++;

				/* labels have to have print something */
				if (!fields[i].label) {
					ret = -EINVAL;
					goto err;
				}
				break;
		}
	}

	if (!form->nfields) {
		ret = -EINVAL;
		goto err; /* need at least one field */
	}

	form->fields = mem_reallocarray(NULL, form->nfields,
					sizeof(struct xfield));
	form->labels = mem_reallocarray(NULL, form->nlabels,
					sizeof(struct xfield));
	if (!form->fields || !form->labels) {
		ret = -ENOMEM;
		goto err_free;
	}

	/* duplicate field descriptions */
	for (i = 0, fi = 0, li = 0; i < nfields; i++) {
		const struct xfield_def *src = &fields[i];
		struct xfield *dst = NULL;
		struct buffer *buf = NULL;
		bool field;

		switch (src->type) {
			case XFIELD_TYPE_FIELD:
				dst = &form->fields[fi++];
				field = true;

				buf = buffer_alloc(src->width * 2);
				if (IS_ERR(buf)) {
					ret = PTR_ERR(buf);
					goto err_fields;
				}
				break;
			case XFIELD_TYPE_LABEL:
				dst = &form->labels[li++];
				field = false;
				break;
		}

		ASSERT3P(dst, !=, NULL);

		dst->def.type = src->type;
		dst->def.row = src->row;
		dst->def.col = src->col;
		dst->def.width = src->width;
		dst->def.label = str_getref(src->label);
		dst->def.attr = src->attr;
		dst->def.private = src->private;
		dst->form = form;
		dst->buf = buf;
		dst->cursor.off = 0; /* beginning of field */
		dst->cursor.scroll = 0; /* beginning of field */
	}

	form->curfield = 0; /* first field */

	keypad(form->formwin, TRUE);

	draw_labels(form);
	draw_fields(form);

	return 0;

err_fields:
	for (i = 0; i < fi; i++) {
		ASSERT3P(form->fields[i].def.label, ==, NULL);
		buffer_free(form->fields[i].buf);
	}

	for (i = 0; i < li; i++) {
		str_putref(form->labels[i].def.label);
		ASSERT3P(form->labels[i].buf, ==, NULL);
	}

err_free:
	free(form->fields);
	free(form->labels);

err:
	delwin(form->basewin);
	delwin(form->formwin);

	return ret;
}

void xform_free(struct xform *form)
{
	size_t i;

	/*
	 * Cleanup form
	 */
	for (i = 0; i < form->nfields; i++) {
		ASSERT3P(form->fields[i].def.label, ==, NULL);
		buffer_free(form->fields[i].buf);
	}

	for (i = 0; i < form->nlabels; i++) {
		str_putref(form->labels[i].def.label);
		ASSERT3P(form->labels[i].buf, ==, NULL);
	}

	free(form->fields);
	free(form->labels);

	/*
	 * Cleanup windows
	 */
	delwin(form->formwin);
	delwin(form->basewin);
}

const void *xform_field_get(struct xform *form, int idx, size_t *size_r)
{
	struct xfield *field = canonical_field(form, idx);

	*size_r = buffer_size(field->buf);

	return buffer_data(field->buf);
}

void *xform_field_get_private(struct xform *form, int idx)
{
	struct xfield *field = canonical_field(form, idx);

	return field->def.private;
}

void xform_field_set(struct xform *form, int idx, const char *data)
{
	struct xfield *field = canonical_field(form, idx);
	ssize_t len = strlen(data);

	ASSERT0(buffer_truncate(field->buf, len));

	ASSERT3S(buffer_pwrite(field->buf, data, len, 0), ==, len);

	/*
	 * we just display the beginning because it is easier than trying to
	 * figure out where the cursor should be
	 */
	field->cursor.off = 0;
	field->cursor.scroll = 0;

	draw_field(field);
}

void xform_field_set_private(struct xform *form, int idx, void *private)
{
	struct xfield *field = canonical_field(form, idx);

	field->def.private = private;
}

void xform_field_clear(struct xform *form, int idx)
{
	struct xfield *field = canonical_field(form, idx);

	ASSERT0(buffer_truncate(field->buf, 0));

	/* nothing to display => move cursor to the beginning of the field */
	field->cursor.off = 0;
	field->cursor.scroll = 0;

	draw_field(field);
}

static int do_char_del(struct xfield *field, bool prev)
{
	size_t off = field->cursor.scroll + field->cursor.off;
	int ret;

	if ((prev && !off) ||
	    (!prev && (off == buffer_size(field->buf)))) {
		/* nothing to delete, but that's not an error */
		beep();
		return 0; /* nothing to delete, but still ok */
	}

	if (prev)
		off--;

	ret = buffer_delete(field->buf, off, 1); /* FIXME: handle utf8 */
	if (ret)
		return ret;

	if (prev) {
		if (field->cursor.scroll &&
		    (field->cursor.off == (field->def.width - 1)))
			field->cursor.scroll--;
		else
			field->cursor.off--;
	}

	draw_field(field);

	return 0;
}

int xform_driver(struct xform *form, int req)
{
	struct xfield *field = &form->fields[form->curfield];

	if (req >= 0) {
		int ret;

		/* FIXME: handle utf8 */
		ret = buffer_insert_c(field->buf, req,
				      field->cursor.scroll + field->cursor.off);
		if (ret)
			return ret;

		/* advance the cursor, and scroll if necessary */
		field->cursor.off++;
		if (field->cursor.off >= field->def.width) {
			field->cursor.off = field->def.width - 1;
			field->cursor.scroll++;
		}

		draw_field(field);

		return 0;
	} else {
		switch ((enum xform_req) req) {
			case XFORM_REQ_FIRST_FIELD:
				form->curfield = 0;
				return 0;
			case XFORM_REQ_NEXT_FIELD:
				form->curfield++;
				if (form->curfield >= form->nfields)
					form->curfield = 0;
				return 0;
			case XFORM_REQ_BEG_FIELD:
				field->cursor.off = 0;
				field->cursor.scroll = 0;
				draw_field(field);
				return 0;
			case XFORM_REQ_END_FIELD: {
				struct xfield *f = field; /* shorter name */
				size_t size = buffer_size(f->buf);
				size_t width = f->def.width;

				if (size >= width) {
					f->cursor.off = width - 1;
					f->cursor.scroll = size - width + 1;
				} else {
					f->cursor.off = size;
					f->cursor.scroll = 0;
				}

				draw_field(f);

				return 0;
			}
			case XFORM_REQ_PREV_CHAR:
				/* move cursor back and scroll if needed */
				if (field->cursor.off) {
					field->cursor.off--;
				} else if (field->cursor.scroll) {
					field->cursor.scroll--;
					draw_field(field);
				} else {
					beep();
				}

				return 0;
			case XFORM_REQ_NEXT_CHAR: {
				/* move cursor forward and scroll if needed */
				struct xfield *f = field; /* shorter name */
				size_t bufoff = f->cursor.scroll;
				size_t bufrem = buffer_size(f->buf) - bufoff;
				size_t bufvis = MIN(bufrem, f->def.width);
				size_t curoff = f->cursor.off;
				bool lastcol = (curoff == (f->def.width - 1));
				bool more = (bufrem >= f->def.width);

				if (!lastcol && (curoff < bufvis)) {
					f->cursor.off++;
				} else if (lastcol && more) {
					f->cursor.scroll++;
					draw_field(f);
				} else {
					beep();
				}
				return 0;
			}
			case XFORM_REQ_DEL_PREV:
				return do_char_del(field, true);
			case XFORM_REQ_DEL_THIS:
				return do_char_del(field, false);
		}

		return -EINVAL; /* unknown request */
	}
}

void xform_cursor_warp(struct xform *form)
{
	struct xfield *f = canonical_field(form, -1);
	int r, c;

	r = f->def.row;
	c = f->def.col + f->cursor.off;

	wmove(form->formwin, r, c);
}
