/*
 * Copyright (c) 2020-2022 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/mem.h>

#include <hlog-lua/xlua.h>

#include "script.h"

#define EVENT_QSO_INIT		"hlog-event-qso-init"
#define EVENT_QSO_DONE		"hlog-event-qso-done"
#define EVENT_FIELD_CHANGED	"hlog-event-field-changed"

static int __append_field(struct xfield_def **fields, size_t *nfields,
			  struct xfield_def *tmpl)
{
	struct xfield_def *tmp;

	tmp = mem_reallocarray(*fields, *nfields + 1, sizeof(struct xfield_def));
	if (!tmp)
		return -ENOMEM;

	tmp[*nfields] = *tmpl;

	*fields = tmp;
	(*nfields)++;

	return 0;
}

static int __load_labels(lua_State *L, struct xfield_def **fields,
			 size_t *nfields)
{
	int ret;

	/* stack: ..., table (top) */

	/* extract the 3 fields */
	lua_geti(L, -1, 1);
	lua_geti(L, -2, 2);
	lua_geti(L, -3, 3);

	/* type check */
	if (!lua_isinteger(L, -3) || !lua_isinteger(L, -2) ||
	    !lua_isstring(L, -1)) {
		fprintf(stderr, "Error: 'labels' values must be "
			"{ row (int), col (int), label (string) }, "
			"found { %s, %s, %s }\n",
			lua_typename(L, lua_type(L, -3)),
			lua_typename(L, lua_type(L, -2)),
			lua_typename(L, lua_type(L, -1)));
		ret = -EINVAL;
	} else {
		struct xfield_def tmpl = {
			.type = XFIELD_TYPE_LABEL,
			.row = lua_tointeger(L, -3),
			.col = lua_tointeger(L, -2),
			.width = strlen(lua_tostring(L, -1)),
			.label = STR_DUP(lua_tostring(L, -1)),
		};

		ret = __append_field(fields, nfields, &tmpl);
	}

	/* pop the fields */
	lua_pop(L, 3);

	return ret;
}

static int __load_fields(lua_State *L, struct xfield_def **fields,
			 size_t *nfields)
{
	int ret;

	/* stack: ..., table (top) */

	/* extract the 4 fields */
	lua_geti(L, -1, 1);
	lua_geti(L, -2, 2);
	lua_geti(L, -3, 3);
	lua_geti(L, -4, 4);

	/* type check */
	if (!lua_isinteger(L, -4) || !lua_isinteger(L, -3) ||
	    !lua_isinteger(L, -2) || !lua_isstring(L, -1)) {
		fprintf(stderr, "Error: 'fields' values must be "
			"{ row (int), col (int), width (int) label (string) }, "
			"found { %s, %s, %s, %s }\n",
			lua_typename(L, lua_type(L, -4)),
			lua_typename(L, lua_type(L, -3)),
			lua_typename(L, lua_type(L, -2)),
			lua_typename(L, lua_type(L, -1)));
		ret = -EINVAL;
	} else {
		struct xfield_def tmpl = {
			.type = XFIELD_TYPE_FIELD,
			.row = lua_tointeger(L, -4),
			.col = lua_tointeger(L, -3),
			.width = lua_tointeger(L, -2),
			.private = STR_DUP(lua_tostring(L, -1)),
		};

		ret = __append_field(fields, nfields, &tmpl);
	}

	/* pop the fields */
	lua_pop(L, 4);

	return ret;
}

/* expects the script definition table at the top of the stack */
static int load_labels_or_fields(lua_State *L, const char *table_name,
				 struct xfield_def **fields, size_t *nfields,
				 int (*load)(lua_State *, struct xfield_def **,
					     size_t *))
{
	lua_Integer i;
	int type;
	int ret;

	type = lua_getfield(L, -1, table_name);
	if (type != LUA_TTABLE) {
		fprintf(stderr, "Error: need '%s' sub-table, got %s\n",
			table_name, lua_typename(L, type));
		lua_pop(L, 1); /* pop whatever we just got */
		return -EINVAL;
	}

	/* iterate over the table, each value is a table with 3 or 4 items */
	for (i = 1;; i++) {
		type = lua_geti(L, -1, i);
		if (type == LUA_TNIL) {
			lua_pop(L, 1); /* pop the nil */
			break;
		}

		/* stack: script def table, @table_name table, value (top) */

		ret = load(L, fields, nfields);

		/* pop the value table */
		lua_pop(L, 1);

		if (ret)
			goto err;
	}

	ret = 0;

err:
	/* pop the @table_name table */
	lua_pop(L, 1);

	ASSERT_LUA_STACK(L, 1);

	return ret;
}

/* expects the script definition table at the top of the stack */
static int load_event_handlers(lua_State *L)
{
	int type;
	int ret;

	type = lua_getfield(L, -1, "events");
	if (type != LUA_TTABLE) {
		fprintf(stderr, "Error: need 'events' sub-table, got %s\n",
			lua_typename(L, type));
		lua_pop(L, 1); /* pop whatever we just got */
		return -EINVAL;
	}

	ret = 0;

	lua_pushnil(L);
	while (lua_next(L, -2)) {
		/* stack: script def table, events table, key, value */

		/* type check */
		if (!lua_isstring(L, -2) || !lua_isfunction(L, -1)) {
			fprintf(stderr, "Error: 'events' values must be "
				"string=function, got %s=%s\n",
				lua_typename(L, lua_type(L, -2)),
				lua_typename(L, lua_type(L, -1)));
			ret = -EINVAL;
		} else if (!strcmp(lua_tostring(L, -2), "qso_init")) {
			lua_pushvalue(L, -1); /* duplicate the function */
			lua_setfield(L, LUA_REGISTRYINDEX, EVENT_QSO_INIT);
		} else if (!strcmp(lua_tostring(L, -2), "qso_done")) {
			lua_pushvalue(L, -1); /* duplicate the function */
			lua_setfield(L, LUA_REGISTRYINDEX, EVENT_QSO_DONE);
		} else if (!strcmp(lua_tostring(L, -2), "field_changed")) {
			lua_pushvalue(L, -1); /* duplicate the function */
			lua_setfield(L, LUA_REGISTRYINDEX, EVENT_FIELD_CHANGED);
		} else {
			fprintf(stderr, "Warning: unknown event '%s' - "
				"ignored\n", lua_tostring(L, -2));
		}

		/* pop the value */
		lua_pop(L, 1);
	}

	/* pop the events table */
	lua_pop(L, 1);

	ASSERT_LUA_STACK(L, 1);

	return ret;
}

int script_exists(const struct contest_params *params)
{
	char file[128];

	if (!params->name)
		return -EINVAL;

	snprintf(file, sizeof(file), "contests/%s.lua", params->name);

	return xlua_script_find(file, NULL, NULL);
}

lua_State *script_load(struct xfield_def **fields, size_t *nfields,
		       const struct contest_params *params)
{
	char error[128];
	char file[128];
	lua_State *L;
	int ret;

	if (!params->name)
		return ERR_PTR(-EINVAL);

	snprintf(file, sizeof(file), "contests/%s.lua", params->name);

	L = xlua_init_script(file, 1, error, sizeof(error));
	if (IS_ERR(L)) {
		printf("Error: failed to initialize lua: %s\n", error);
		return L;
	}

	if (lua_type(L, -1) != LUA_TTABLE) {
		fprintf(stderr, "Error: contest script init failed: "
			"expected table, got %s\n",
			lua_typename(L, lua_type(L, -1)));
		ret = -EINVAL;
		goto err;
	}

	/* set up labels */
	ret = load_labels_or_fields(L, "labels", fields, nfields,
				    __load_labels);
	if (ret)
		goto err;

	/* set up fields */
	ret = load_labels_or_fields(L, "fields", fields, nfields,
				    __load_fields);
	if (ret)
		goto err;

	/* sanity check event handlers */
	ret = load_event_handlers(L);
	if (ret)
		goto err;

	/* TODO: add contest fields to qso-line (or maybe replace instead?) */

	lua_pop(L, 1); /* pop the script definition table */

	ASSERT_LUA_STACK(L, 0);

	return L;

err:
	lua_close(L);

	return ERR_PTR(ret);
}

int script_event_field_change(lua_State *L, struct qso *qso,
			      struct xform *form,
			      char *error, size_t errlen,
			      void (*cb)(struct xform *, const char *,
					 const char *))
{
	size_t i;
	int type;
	int ret;

	type = lua_getfield(L, LUA_REGISTRYINDEX, EVENT_FIELD_CHANGED);
	if (type == LUA_TNIL) {
		lua_pop(L, 1);
		ASSERT_LUA_STACK(L, 0);
		return 0;
	}

	/*
	 * call the function with args: qso, fields, changed-label
	 */

	/* push arg: qso */
	ret = xlua_pushqso(L, qso_getref(qso), false, error, errlen);
	if (ret) {
		lua_pop(L, 1); /* pop the function */
		goto out;
	}

	/* push arg: fields table */
	lua_newtable(L);

	for (i = 0; i < form->nfields; i++) {
		const void *data;
		size_t size;

		data = xform_field_get(form, i, &size);

		lua_pushlstring(L, data, size);
		lua_setfield(L, -2, str_cstr(xform_field_get_private(form, i)));
	}

	/* push arg: changed-label */
	lua_pushstring(L, str_cstr(xform_field_get_private(form, -1)));

	/* call */
	ret = xlua_call(L, NULL, 3, 1, error, errlen);
	if (ret)
		goto out;

	/*
	 * process the result
	 */

	if (lua_istable(L, -1)) {
		/* process the table */

		/* for each key-value in the table, invoke the callback */
		lua_pushnil(L);
		while (lua_next(L, -2)) {
			/* stack: table, key, value (top) */

			if (lua_isstring(L, -2) && lua_isstring(L, -1))
				cb(form, lua_tostring(L, -2),
				   lua_tostring(L, -1));

			/* pop the value */
			lua_pop(L, 1);
		}
	} else if (!lua_isnil(L, -1)) {
		/* display a message but don't fail */
		strcpy_safe(error, "field-changed method returned "
			    "non-table", errlen);
	}

	lua_pop(L, 1); /* pop the result */

out:
	ASSERT_LUA_STACK(L, 0);

	return ret;
}

static int event_qso(lua_State *L, struct qso *qso, const char *event,
		     char *error, size_t errlen)
{
	int type;
	int ret;

	type = lua_getfield(L, LUA_REGISTRYINDEX, event);
	if (type == LUA_TNIL) {
		lua_pop(L, 1);
		ASSERT_LUA_STACK(L, 0);
		return 0;
	}

	ret = xlua_qso_call(L, NULL, qso_getref(qso), false, error, errlen);

	ASSERT_LUA_STACK(L, 0);

	return ret;
}

int script_event_qso_init(lua_State *L, struct qso *qso,
			  char *error, size_t errlen)
{
	return event_qso(L, qso, EVENT_QSO_INIT, error, errlen);
}

int script_event_qso_done(lua_State *L, struct qso *qso,
			  char *error, size_t errlen)
{
	return event_qso(L, qso, EVENT_QSO_DONE, error, errlen);
}
