/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <zmq.h>

#include <jeffpc/error.h>
#include <jeffpc/types.h>
#include <jeffpc/cstr.h>

#include <hlog-rpc/announce.h>
#include <hlog-rpc/announce-pub.h>

static void *sock;
static void *ctx;

int announce_start(void)
{
	int ret;

	ctx = zmq_ctx_new();
	if (!ctx) {
		cmn_err(CE_ERROR, "failed to create zmq context: %s",
			strerror(errno));
		return -errno;
	}

	sock = zmq_socket(ctx, ZMQ_PUB);
	if (!sock) {
		cmn_err(CE_ERROR, "failed to create zmq socket: %s",
			strerror(errno));
		ret = -errno;
		goto err;
	}

	if (zmq_bind(sock, ANNOUNCE_SOCK_NAME)) {
		cmn_err(CE_ERROR, "failed to bind zmq socket: %s",
			strerror(errno));
		ret = -errno;
		goto err_sock;
	}

	cmn_err(CE_INFO, "Bound to: %s (MQ)", ANNOUNCE_SOCK_NAME);

	return 0;

err_sock:
	if (zmq_close(sock))
		cmn_err(CE_ERROR, "failed to close zmq socket: %s",
			strerror(errno));

err:
	if (zmq_ctx_destroy(ctx))
		cmn_err(CE_ERROR, "failed to destroy zmq context: %s",
			strerror(errno));

	return ret;
}

void announce_stop(void)
{
	if (zmq_close(sock))
		cmn_err(CE_ERROR, "failed to close zmq socket: %s",
			strerror(errno));

	if (zmq_ctx_destroy(ctx))
		cmn_err(CE_ERROR, "failed to destroy zmq context: %s",
			strerror(errno));
}

void announce_publish(struct announce_msg *msg)
{
	int ret;

	/*
	 * Don't trust that padding is zero filled
	 */
	msg->_pad0 = 0;
	msg->_pad1 = 0;

	switch ((enum announce_msg_type) be8_to_cpu(msg->major)) {
		case ANNOUNCE_QSO:
			memset(msg->u.qso._pad2, 0, sizeof(msg->u.qso._pad2));
			break;
		case ANNOUNCE_LOC:
			memset(msg->u.loc._pad2, 0, sizeof(msg->u.loc._pad2));
			break;
		case ANNOUNCE_RIG:
			memset(msg->u.rig._pad2, 0, sizeof(msg->u.rig._pad2));
			break;
	}

	/*
	 * Publish to the MQ
	 */
	ret = zmq_send(sock, msg, sizeof(*msg), 0);
	if (ret < 0)
		cmn_err(CE_WARN, "announce: failed to send notification: %s",
			strerror(errno));
	else if (ret != sizeof(*msg))
		cmn_err(CE_WARN, "announce: failed to send the entire "
			"notification of %zu bytes (%d sent)",
			sizeof(*msg), ret);
}
