/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <zmq.h>

#include <jeffpc/error.h>
#include <jeffpc/types.h>
#include <jeffpc/thread.h>

#include <hlog-rpc/announce.h>
#include <hlog-rpc/announce-sub.h>

static pthread_t announce_thread;
static void *ctx;
static void *sock;

static struct {
	void (*qso)(enum announce_qso_msg_type, struct xuuid *);
	void (*loc)(enum announce_loc_msg_type, int mode, const char *grid);
	void (*rig)(enum announce_rig_msg_type, uint64_t tx_freq,
		    uint64_t rx_freq, uint32_t power);
} notify;

static void interpret_msg(struct announce_msg *msg)
{
	/* byte-order */
	msg->major = be8_to_cpu(msg->major);
	msg->minor = be8_to_cpu(msg->minor);

	switch ((enum announce_msg_type) msg->major) {
		case ANNOUNCE_QSO:
			break; /* nothing else to byte-order */
		case ANNOUNCE_LOC:
			msg->u.loc.mode = be8_to_cpu(msg->u.loc.mode);
			break;
		case ANNOUNCE_RIG:
			msg->u.rig.tx_freq = be64_to_cpu(msg->u.rig.tx_freq);
			msg->u.rig.rx_freq = be64_to_cpu(msg->u.rig.rx_freq);
			msg->u.rig.power = be32_to_cpu(msg->u.rig.power);
			break;
	}

	/* actually interpret */
	switch ((enum announce_msg_type) msg->major) {
		case ANNOUNCE_QSO:
			if (notify.qso) {
				struct xuuid uuid;

				memcpy(uuid.raw, msg->u.qso.uuid, sizeof(uuid));

				notify.qso(msg->minor, &uuid);
			}
			break;
		case ANNOUNCE_LOC:
			if (notify.loc)
				notify.loc(msg->minor, msg->u.loc.mode,
					   (const char *) msg->u.loc.grid);
			break;
		case ANNOUNCE_RIG:
			if (notify.rig)
				notify.rig(msg->minor, msg->u.rig.tx_freq,
					   msg->u.rig.rx_freq,
					   msg->u.rig.power);
			break;
	}
}

static void *process(void *unused)
{
	struct announce_msg msg;
	int ret;

	for (;;) {
		ret = zmq_recv(sock, &msg, sizeof(struct announce_msg), 0);
		if (ret < 0) {
			/* error - ignore it */
			continue;
		}

		if (ret != sizeof(struct announce_msg)) {
			/* truncated message - ignore it */
			continue;
		}

		interpret_msg(&msg);
	}

	return NULL;
}

int announce_connect(void (*qso)(enum announce_qso_msg_type, struct xuuid *),
		     void (*loc)(enum announce_loc_msg_type, int mode,
				 const char *grid),
		     void (*rig)(enum announce_rig_msg_type, uint64_t tx_freq,
				 uint64_t rx_freq, uint32_t power))
{
	int ret;

	notify.qso = qso;
	notify.loc = loc;
	notify.rig = rig;

	ctx = zmq_ctx_new();
	if (!ctx)
		return -errno;

	sock = zmq_socket(ctx, ZMQ_SUB);
	if (!sock) {
		ret = -errno;
		goto err;
	}

	ret = zmq_setsockopt(sock, ZMQ_SUBSCRIBE, NULL, 0);
	if (ret) {
		ret = -errno;
		goto err_sock;
	}

	ret = zmq_connect(sock, ANNOUNCE_SOCK_NAME);
	if (ret) {
		ret = -errno;
		goto err_sock;
	}

	ret = xthr_create_named(&announce_thread, "announce", process, NULL);
	if (ret)
		goto err_sock;

	return 0;

err_sock:
	if (zmq_close(sock))
		fprintf(stderr, "Error: failed to close socket: %s\n",
			strerror(errno));

err:
	if (zmq_ctx_destroy(ctx))
		fprintf(stderr, "Error: failed to destroy context: %s\n",
			strerror(errno));

	return ret;
}
