/*
 * Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/nvl.h>

#include "xlua_impl.h"
#include "wrapper.h"

static int xlua_nvl_index(lua_State *L);
static int xlua_nvl_newindex(lua_State *L);

static const luaL_Reg __nvl_fxns[] = {
	{ "__index", xlua_nvl_index },
	{ "__newindex", xlua_nvl_newindex },
	{ NULL, NULL },
};

WRAP_C_DATA(nvl, struct nvlist, nvl_putref, __nvl_fxns)

/* handle a nvl[key], where nvlist is -2 and key is -1 [-2,+1,e] */
static int xlua_nvl_index(lua_State *L)
{
	struct nvlist *nvl;
	const char *key;
	struct val *val;
	bool ro;

	xlua_checkstack(L, 2, 2, 1);

	nvl = xlua_nvl_getptr(L, -2, &ro);
	key = luaL_checkstring(L, -1);

	val = nvl_lookup_val(nvl, key);
	if (IS_ERR(val)) {
		int ret = PTR_ERR(val);

		if (ret != -ENOENT)
			return luaL_error(L, "failed to lookup key in "
					  "nvlist: %s", xstrerror(ret));

		/* not found ==> nil */
		lua_pushnil(L);
		return 1;
	}

	switch (val->type) {
		case VT_NULL:
			lua_pushnil(L);
			goto out;
		case VT_INT:
			lua_pushinteger(L, val->i);
			goto out;
		case VT_STR:
			lua_pushstring(L, val_cstr(val));
			goto out;
		case VT_BOOL:
			lua_pushboolean(L, val->b);
			goto out;
		case VT_BLOB:
			lua_pushlstring(L, val->blob.ptr,
					val->blob.size);
			goto out;
		case VT_NVL:
			xlua_pushnvl_throws(L, val_cast_to_nvl(val), ro);
			val = NULL; /* avoid double-putref */
			goto out;
		case VT_SYM:
		case VT_CONS:
		case VT_CHAR:
		case VT_ARRAY: {
			enum val_type type = val->type;

			val_putref(val);
			return luaL_error(L, "unhandled val type: %s",
					  val_typename(type));
		}
	}

	panic("unknown val type %d", val->type);

out:
	val_putref(val);

	return 1;
}

/*
 * handle a nvlist[key] = value, where nvlist is -3, key is -2, and value is
 * -1 [-3,+0,e]
 */
static int xlua_nvl_newindex(lua_State *L)
{
	struct nvlist *nvl;
	const char *key;
	struct val *val;
	int ret;
	bool ro;

	xlua_checkstack(L, 3, 3, 0);

	nvl = xlua_nvl_getptr(L, -3, &ro);
	key = luaL_checkstring(L, -2);

	if (ro)
		return luaL_error(L, "cannot change read-only nvlist");

	switch (lua_type(L, -1)) {
		case LUA_TNIL:
			goto unset;
		case LUA_TSTRING: {
			const char *cstr;
			size_t len;

			cstr = lua_tolstring(L, -1, &len);

			if (len != strlen(cstr))
				val = val_alloc_blob_dup(cstr, len);
			else
				val = val_dup_str(cstr);

			if (IS_ERR(val))
				return luaL_error(L, "failed to convert lua "
						  "string to nvlist item: %s",
						  xstrerror(PTR_ERR(val)));

			goto set;
		}
		case LUA_TNUMBER: {
			lua_Integer i;
			int isnum;

			i = lua_tointegerx(L, -1, &isnum);
			if (!isnum)
				return luaL_error(L, "failed to convert lua "
						  "number to nvlist item");

			val = val_alloc_int(i);
			if (IS_ERR(val))
				return luaL_error(L, "failed to convert lua "
						  "number to nvlist item: %s",
						  xstrerror(PTR_ERR(val)));

			goto set;
		}
	}

	return luaL_error(L, "can't set nvlist value to lua %s type",
			  lua_typename(L, lua_type(L, -1)));

set:
	ret = nvl_set(nvl, key, val);
	if (ret)
		return luaL_error(L, "failed to set nvlist value: %s",
				  xstrerror(ret));

	return 0;

unset:
	ret = nvl_unset(nvl, key);
	if (ret && (ret != -ENOENT))
		return luaL_error(L, "failed to unset nvlist value: %s",
				  xstrerror(ret));

	return 0;
}

/*
 * hlog.nvl.clear() [-1,+0,e]
 */
static int xlua_nvl_clear(lua_State *L)
{
	struct nvlist *nvl;
	bool ro;

	xlua_checkstack(L, 1, 1, 0);

	nvl = xlua_nvl_getptr(L, -1, &ro);

	if (ro)
		return luaL_error(L, "cannot clear a read-only nvlist");

	nvl_clear(nvl);

	return 0;
}

static const luaL_Reg nvl_funcs[] = {
	{ "clear", xlua_nvl_clear },
	{ NULL, NULL },
};

void xlua_pushlib_nvl(lua_State *L)
{
	luaL_newlib(L, nvl_funcs);
}
