--
-- Copyright (c) 2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
--
-- Permission is hereby granted, free of charge, to any person obtaining a copy
-- of this software and associated documentation files (the "Software"), to deal
-- in the Software without restriction, including without limitation the rights
-- to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-- copies of the Software, and to permit persons to whom the Software is
-- furnished to do so, subject to the following conditions:
--
-- The above copyright notice and this permission notice shall be included in
-- all copies or substantial portions of the Software.
--
-- THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-- IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-- FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-- AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-- LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-- OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
-- SOFTWARE.
--

local function copy_if_not_nil(dst, dkey, src)
	if src ~= nil then
		dst[dkey] = src
	end
end

-- copy country data into QSO
--
-- If the station callsign is nil (i.e., empty), we clear the various
-- QSO location fields.  In theory, we could clear the fields whenever we
-- get a lookup failure, but that seems a bit dangerous.
local function sync_zones(qso)
	local call = qso.rx.station_call
	local state = qso.rx.state

	if call ~= nil then
		local cty = hlog.country.lookup(call, state)
		if cty == nil then
			return
		end

		copy_if_not_nil(qso.rx, 'dxcc', cty.dxcc)
		copy_if_not_nil(qso.rx, 'itu', cty.itu)
		copy_if_not_nil(qso.rx, 'cqz', cty.cqz)
		copy_if_not_nil(qso.rx, 'country', cty.country)
		copy_if_not_nil(qso.rx, 'continent', cty.continent)
	else
		qso.rx.dxcc = nil
		qso.rx.itu = nil
		qso.rx.cqz = nil
		qso.rx.country = nil
		qso.rx.continent = nil
	end
end

local function field_changed(qso, fields, field)
	local value = fields[field]

	if field == "rx.station_call" then
		qso.rx.station_call = value
		sync_zones(qso)
	elseif field == "tx.rst" then
		qso.tx.rst = value
	elseif field == "rx.rst" then
		qso.rx.rst = value
	elseif field == "rx.state" then
		qso.rx.state = value:upper()
		sync_zones(qso)
	elseif field == "name" then
		qso.rx.name = value
	elseif field == "ref" then
		qso.additional.rx_pota = value:upper()
		-- TODO: do a ref lookup and fill in any info we may be
		-- missing (qth, state, cqz, ituz, ...)
	else
		error(string.format("Unknown field '%s' encountered", field))
	end
end

return {
	labels = {
		{ 0,  0, "St. Call" },
		{ 0, 11, "RSTs" },
		{ 0, 16, "RSTr" },
		{ 0, 21, "St" },
		{ 0, 24, "Name" },
		{ 0, 41, "Ref" },
	},

	fields = {
		{ 1,  0, 10, "rx.station_call", },
		{ 1, 11,  4, "tx.rst", },
		{ 1, 16,  4, "rx.rst", },
		{ 1, 21,  2, "rx.state", },
		{ 1, 24, 16, "name", },
		{ 1, 41,  6, "ref", },
	},

	events = {
		-- called when QSO is being initialized
		--
		-- We rely on the tx_{pota,wwff} additional fields already
		-- being set by the template.
		qso_init = function(qso)
			local ref = qso.additional['tx_pota']

			contest.lcd(ref ~= nil and ref or "?")
		end,

		-- called when QSO is complete & saved, just prior to a reset
		qso_done = function(qso)
			-- TODO: update dup info
		end,

		-- called whenever a contest field changes
		field_changed = field_changed,
	},
}
