/*
 * Copyright (c) 2020-2021 Josef 'Jeff' Sipek <jeffpc@josefsipek.net>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <jeffpc/cstr.h>
#include <jeffpc/io.h>

#include "xlua_impl.h"
#include "script.h"

int lua2errno(int e)
{
	switch (e) {
		case LUA_OK:
			return 0;
		case LUA_ERRFILE: /* file-related error */
			return -EIO;
		case LUA_ERRSYNTAX: /* syntax error during precompilation */
		case LUA_ERRRUN: /* runtime error */
		case LUA_ERRERR: /* msg handler execution error */
			return -EINVAL;
		case LUA_ERRMEM: /* out of memory */
#ifdef LUA_ERRGCMM
		case LUA_ERRGCMM: /* __gc metamethod error */
#endif
			return -ENOMEM;
	}

	panic("unhandled lua error %d", e);
}

/* push the current time [-0,+1,-] */
static int xlua_gettime(lua_State *L)
{
	lua_pushinteger(L, gettime());

	return 1;
}

static const luaL_Reg xlua_funcs[] = {
	{ "gettime", xlua_gettime },
	{ NULL, NULL },
};

static void xlua_register(lua_State *L)
{
	lua_newtable(L);

	lua_pushliteral(L, "adif");
	xlua_pushlib_adif(L);
	lua_settable(L, -3);

	lua_pushliteral(L, "country");
	xlua_pushlib_country(L);
	lua_settable(L, -3);

	lua_pushliteral(L, "index");
	xlua_pushlib_index(L);
	lua_settable(L, -3);

	lua_pushliteral(L, "nvl");
	xlua_pushlib_nvl(L);
	lua_settable(L, -3);

	lua_pushliteral(L, "qso");
	xlua_pushlib_qso(L);
	lua_settable(L, -3);

	luaL_setfuncs(L, xlua_funcs, 0);

	lua_setglobal(L, "hlog");
}

/* read a chunk of data */
static const char *xlua_init_reader(lua_State *L, void *data, size_t *size)
{
	struct buffer *buf = data;
	const void *ptr;

	if (!buffer_remain(buf))
		return NULL; /* EOF */

	*size = buffer_remain(buf);
	ptr = buffer_data_current(buf);

	/* advance to the EOF */
	ASSERT3S(buffer_seek(buf, 0, SEEK_END), >=, 0);

	return ptr;
}

static lua_State *__xlua_init(const char *name, const uint8_t *data,
			      size_t size, int nresults, char *error,
			      size_t errlen)
{
	struct buffer buf;
	lua_State *L;
	int ret;

	buffer_init_static(&buf, data, size, size, false);

	L = luaL_newstate();
	if (!L) {
		strcpy_safe(error, "failed to allocate lua state", errlen);
		return ERR_PTR(-ENOMEM);
	}

	luaL_openlibs(L);

	xlua_register(L);

	ret = lua_load(L, xlua_init_reader, &buf, name, NULL);
	if (ret != LUA_OK)
		goto err;

	ret = lua_pcall(L, 0, nresults, 0);
	if (ret)
		goto err;

	ASSERT_LUA_STACK(L, nresults);

	return L;

err:
	COPY_ERROR(L, -1, error, errlen);

	lua_close(L);

	return ERR_PTR(lua2errno(ret));
}

lua_State *xlua_init_at(int dirfd, const char *fname, int nresults, char *error,
			size_t errlen)
{
	lua_State *L;
	char *raw;
	size_t size;

	if (!fname || (nresults < 0)) {
		strcpy_safe(error, "invalid argument", errlen);
		return ERR_PTR(-EINVAL);
	}

	raw = read_file_len_at(dirfd, fname, &size);
	if (IS_ERR(raw)) {
		strcpy_safe(error, xstrerror(PTR_ERR(raw)), errlen);
		return ERR_CAST(raw);
	}

	L = __xlua_init(fname, (const uint8_t *) raw, size, nresults, error,
			errlen);

	free(raw);

	return L;
}

lua_State *xlua_init_script(const char *scriptname, int nresults, char *error,
			    size_t errlen)
{
	const uint8_t *data;
	size_t size;
	int ret;

	if (!scriptname || (nresults < 0)) {
		strcpy_safe(error, "invalid argument", errlen);
		return ERR_PTR(-EINVAL);
	}

	ret = xlua_script_find(scriptname, &data, &size);
	if (ret) {
		strcpy_safe(error, xstrerror(ret), errlen);
		return ERR_PTR(ret);
	}

	return __xlua_init(scriptname, data, size, nresults, error, errlen);
}

int xlua_call(lua_State *L, const char *funcname, int nargs, int nresults,
	      char *error, size_t errlen)
{
	const int orig_top = lua_gettop(L) - nargs - (funcname ? 0 : 1);
	int ret;

	ASSERT3S(nresults, >=, 0);

	if (orig_top < 0) {
		lua_pop(L, lua_gettop(L)); /* pop everything */
		ASSERT_LUA_STACK(L, 0);

		strcpy_safe(error, "not enough arguments", errlen);

		return -EINVAL;
	}

	if (funcname) {
		ret = lua_getglobal(L, funcname);

		/* move the function before the args */
		if (nargs)
			lua_insert(L, -(nargs + 1));
	} else {
		ret = lua_type(L, -(nargs + 1));
	}

	if (ret != LUA_TFUNCTION) {
		lua_pop(L, nargs + 1); /* pop func + args */
		ASSERT_LUA_STACK(L, orig_top);

		strcpy_safe(error, "not a function", errlen);

		return -ENOENT;
	}

	ret = lua_pcall(L, nargs, nresults, 0);
	if (ret) {
		COPY_ERROR(L, -1, error, errlen);

		lua_pop(L, 1);
		ASSERT_LUA_STACK(L, orig_top);
	} else {
		ASSERT_LUA_STACK(L, orig_top + nresults);
	}

	return lua2errno(ret);
}

void xlua_dump_stack(lua_State *L)
{
	int top;
	int i;

	top = lua_gettop(L);

	printf("Lua state %p (top = %d)\n", L, top);

	for (i = 1; i <= top; i++) {
		const int t = lua_type(L, i);

		printf("%2d|%3d: ", i, i - top - 1);

		switch (t) {
			case LUA_TSTRING:
				printf("string '%s'", lua_tostring(L, i));
				break;
			case LUA_TBOOLEAN:
				printf("boolean %s", lua_toboolean(L, i) ?
				       "true" : "false");
				break;
			case LUA_TNUMBER:
				printf("number %g", lua_tonumber(L, i));
				break;
			default:
				printf("%s", lua_typename(L, t));
				break;
		}

		printf("\n");
	}

	printf("EOS\n");

	fflush(stdout);
}
